<?php

namespace Drupal\xnttbrapi\Plugin\ExternalEntities\StorageClient;

use Drupal\Core\Form\FormStateInterface;
use Drupal\external_entities\Plugin\ExternalEntities\StorageClient\RestClient;
use Drupal\xnttbrapi\Exception\XnttBrapiException;
use Psr\Http\Message\ResponseInterface;

/**
 * External entities storage client based on a REST API.
 *
 * @StorageClient(
 *   id = "brapi",
 *   label = @Translation("BrAPI"),
 *   description = @Translation("Retrieves external entities from a Breeding API enpoint.")
 * )
 */
class Brapi extends RestClient {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form = parent::buildConfigurationForm($form, $form_state);

    // Add help.
    $form['endpoint']['#title'] = $this->t('BrAPI Call URL');
    $form['endpoint']['#description'] = $this->t(
      'Enter the BrAPI call URL that returns a list of the requested data type. It may contain query filters or can be a search call. Ex.:<ul><li>"https://www.crop-diversity.org/mgis/brapi/v1/germplasm/?instituteCode=TheInstituteName"</li><li>"https://www.crop-diversity.org/mgis5/brapi/v2/search/germplasm/".</li></ul>'
    );
    $form['endpoint']['#attributes']['placeholder'] = $this->t(
      "ex.: https://www.crop-diversity.org/mgis/brapi/v1/germplasm/"
    );
    $form['endpoint']['#required'] = FALSE;

    $form['endpoint_options']['#open'] = TRUE;

    $form['endpoint_options']['cache'] = [
      '#type' => 'hidden',
      '#value' => TRUE,
    ];

    $form['endpoint_options']['single']['#required'] = TRUE;
    $form['endpoint_options']['single']['#title'] = $this->t(
      'BrAPI call to load a single data type by its identifier'
    );
    $form['endpoint_options']['single']['#description'] = $this->t(
      'Enter the BrAPI call URL to use to fetch a single element. You MUST include curly brackets with the BrAPI identifier field name inside. Ex.:<ul><li>"https://www.crop-diversity.org/mgis/brapi/v1/germplasm/{germplasmDbId}"</li><li>"https://www.crop-diversity.org/mgis5/brapi/v2/germplasm?accessionNumber={accessionNumber}".</li></ul>'
    );
    $form['endpoint_options']['single']['#attributes']['placeholder'] = $this->t(
        "ex.: https://www.crop-diversity.org/mgis/brapi/v2/germplasm/{germplasmDbId}"
    );

    $form['endpoint_options']['count'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    $form['endpoint_options']['count_mode'] = [
      '#type' => 'hidden',
      '#value' => 'entities',
    ];

    // Auto-set some REST parameters.
    $form['response_format'] = [
      '#type' => 'hidden',
      '#value' => 'json',
    ];

    unset($form['data_path']['#type']);

    $form['data_path']['list'] = [
      '#type' => 'hidden',
      // Works with both 'result' and 'results'.
      '#value' => '$[result,results].data.*',
    ];

    $form['data_path']['single'] = [
      '#type' => 'hidden',
      '#value' => '$.result',
    ];

    $form['data_path']['keyed_by_id'] = [
      '#type' => 'hidden',
      '#value' => FALSE,
    ];

    $form['data_path']['count'] = [
      '#type' => 'hidden',
      '#value' => '$.metadata.pagination.totalCount',
    ];

    $form['pager']['page_parameter'] = [
      '#type' => 'hidden',
      '#value' => 'page',
    ];

    $form['pager']['page_parameter_type'] = [
      '#type' => 'hidden',
      '#value' => 'pagenum',
    ];

    $form['pager']['page_size_parameter'] = [
      '#type' => 'hidden',
      '#value' => 'pageSize',
    ];

    $form['pager']['page_size_parameter_type'] = [
      '#type' => 'hidden',
      '#value' => 'pagesize',
    ];

    $form['pager']['page_start_one'] = [
      '#type' => 'hidden',
      '#value' => FALSE,
    ];

    // Authentication type set in validation if a key is provided.
    $form['api_key']['type'] = [
      '#type' => 'hidden',
      '#value' => 'none',
    ];

    $form['api_key']['key']['#description'] = $this->t(
      'Enter here the BrAPI token to use if the endpoint requires one.'
    );
    $form['api_key']['key']['#title'] = $this->t('BrAPI Token');

    $form['api_key']['header_name'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    unset($form['http']['#type']);

    $form['http']['headers'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];

    // Add search call body parameters.
    $form['parameters']['#title'] = $this->t('Optional parameters');
    $form['parameters']['#description'] = $this->t(
      'For search calls, optional POST parameters can be specified to pre-filter results. Multiple values can be specified using square brackets and coma-separated values. For instance, use "instituteCode|TheInstituteName" to restrict germplasm to the institute "TheInstituteName" or "species|[SpeciesName1,SpeciesName2]" to restrict to the species "SpeciesName1" or "SpeciesName2".'
    );

    $form['parameters']['list']['#title'] = $this->t('BrAPI search call POST parameters');
    $form['parameters']['list']['#description'] = NULL;
    $form['parameters']['list_param_mode']['#title'] = $this->t('BrAPI Call URL uses');
    $form['parameters']['list_param_mode']['#options'] = [
      'query' => $this->t('GET (list calls)'),
      'body' => $this->t('POST (search calls)'),
    ];
    unset($form['parameters']['list_param_mode']['#description']);

    $form['parameters']['single'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];
    $form['parameters']['single_param_mode'] = [
      '#type' => 'hidden',
      '#value' => 'query',
    ];

    // Filtering.
    unset($form['filtering']['#type']);
    $form['filtering']['basic'] = [
      '#type' => 'hidden',
      // It is patrially TRUE. It could be adjusted later if needed.
      '#value' => TRUE,
    ];
    $form['filtering']['list_support'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];
    $form['filtering']['basic_fields'] = [
      '#type' => 'hidden',
      '#value' => 'none',
    ];
    $form['filtering']['list_join'] = [
      '#type' => 'hidden',
      '#value' => ',',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    // @todo Make sure single endpoint has curly braces.
    // Enable authentication if API key set.
    $api_key = $form_state->getValue('api_key');
    if (!empty($api_key['key'])) {
      // Make sure key type is properly set before parent validation.
      $api_key['type'] = 'bearer';
      $form_state->setValue('api_key', $api_key);
    }
    parent::validateConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    $parameters = $form_state->getValue('parameters') ?? [];
    // Rearrange parameters and handle multiple values.
    foreach (['list', 'single'] as $type) {
      // Make sure the parameters have not been already processed by a child
      // module.
      $value = $parameters[$type] ?? '';
      if (is_string($value)) {
        $lines = explode("\n", $value);
        $lines = array_map('trim', $lines);
        $lines = array_filter($lines, 'strlen');
        $parameters[$type] = [];
        foreach ($lines as $line) {
          $exploded = explode('|', $line);
          if ('' != $exploded[0]) {
            $value = !empty($exploded[1]) ? $exploded[1] : $exploded[0];
            if (('[' == substr($value, 0, 1))
                && (']' == substr($value, -1, 1))
            ) {
              $value = explode(',', substr($value, 1, -1));
            }
            $parameters[$type][$exploded[0]] = $value;
          }
        }
      }
    }
    $form_state->setValue('parameters', $parameters);
    parent::submitConfigurationForm($form, $form_state);
    // @todo Remove missing single endpoint {id} warning.
    // @todo Force identifier field mapping here.
    // $this->externalEntityType->getFieldMapper(); ...
  }

  /**
   * Returns the BrAPI object identifier used.
   *
   * @return string
   *   The BrAPI identifier field name or throws an excetpion.
   *
   * @throw XnttBrapiException if unable to guess object identifier.
   */
  public function getIdentifierFieldName() :string {
    $read_call = $this->getEndpointUrl('read');
    if (!preg_match('#\{(\w+)\}#', $read_call, $matches) || empty($matches[1])) {
      throw new XnttBrapiException(
        "Error: unable to guess identifier field from single data type loading call ($read_call). No identifier field specified in curly braces."
      );
    }
    return $matches[1];
  }

  /**
   * {@inheritdoc}
   */
  public function getEndpointUrl(
    string $request_type = 'list',
    array $parameters = []
  ) :string {
    $endpoint_url = parent::getEndpointUrl($request_type, $parameters);
    $endpoint_url = preg_replace('/\{\w+\}/', '{id}', $endpoint_url);
    return $endpoint_url;
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\xnttbrapi\Exception\XnttBrapiException
   */
  public function getResponseBody(
    ResponseInterface $response,
    string $request_type,
    string $method,
    string $endpoint,
    array $req_parameters = []
  ) :string {

    // Manage asynchronous searches with "searchResultsDbId".
    if (202 == $response->getStatusCode()) {
      // Asynchronous search, get search token.
      $format = $this->configuration['response_format'];
      $decoder = $this->getResponseDecoderFactory()->getDecoder($format);
      // Force string.
      $body = $response->getBody() . '';
      $body = preg_replace('/^\xEF\xBB\xBF/', '', $body);
      $result = $decoder->decode($body);
      $search_results_db_id = $result['result']['searchResultsDbId'] ?? '';
      if (empty($search_results_db_id)) {
        throw new XnttBrapiException(
          "Failed to use BrAPI search call:\nURL: $endpoint\nMethod: $method\nQuery: "
          . print_r($query, TRUE)
          . 'Error: no searchResultsDbId returned while 202 HTTP response code.'
        );
      }
      $method = 'GET';
      $endpoint = rtrim('/', $endpoint) . '/' . $search_results_db_id;
      $request = ['headers' => $request['headers']];
      // Try for 5 min max.
      $timeout = time() + 300;
      $trials = 0;
      // Start to query every 1 sec.
      $sleep_time = 1;
      while ((time() < $timeout)
        && (202 == $response->getStatusCode())
      ) {
        // Wait before re-querying.
        sleep($sleep_time);
        switch (++$trials) {
          case 10:
            // Switch to 2sec delay after 10sec.
            $sleep_time = 2;
            break;

          case 15:
            // Switch to 5sec delay after 20sec.
            $sleep_time = 5;
            break;

          case 29:
            // Switch to 10sec delay after 1min30.
            $sleep_time = 10;
            break;
        }
        // Manage delays between queries.
        $this->ensureQueryLimits();
        $response = $this->httpClient->request(
          $method,
          $endpoint,
          $request
        );
      }
    }

    if (200 == $response->getStatusCode()) {
      // Force string.
      $body = $response->getBody() . '';
      // Remove UTF8 BOM added by some server which may intefer with data
      // parsers.
      $body = preg_replace('/^\xEF\xBB\xBF/', '', $body);
    }
    else {
      // Don't know how to handle non-200 answers: return an empty string.
      $body = '';
    }

    return $body;
  }

}
